/*

*/
package pipeline;
import scene.*;
import framebuffer.*;

/**
    This renderer takes as its input a Scene data structure and
    a FrameBuffer data structure. This renderer mutates the
    FrameBuffer's current viewport so that it is filled in with
    the rendered image of the scene represented by the Scene object.
<p>
    This implements our fourteenth rendering pipeline.
<p>
    Since the model-to-world, world-to-view, normaliztion, and projection
    stages are all implemented by matrix multiplication, this pipeline
    first pre-multiplies those four matrices and then applies them in a
    single pipeline stage, Notice also that the combined transformation
    is "inlined" into this renderer in order to save a method call. And
    in order to save two more method calls, the perspective division and
    viewport transformation are also "inlined" into this renderer
<p>
    The combining and inlining of the several transforms causes a noticeable
    speedup in this renderer. To see the speedup, run the Geometries.java
    program from both this renderer and the previous renderer but in both
    cases, comment out the line of code in Geometries.java that saves the
    FrameBuffer object to the file system. That eliminates from the rendering
    the very unpredictable file system I/O time, and gives a fairly accurate
    comparison of rendering compute time.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;
   public static boolean debug_visible = false;

   /**
      Mutate the FrameBuffer's current viewport so that it holds
      the rendered image of the Scene object.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug || debug_visible )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer a scene without any debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // Create the combined world-to-view, normalize, projection transformation.
      Matrix w2p = scene.camera.projMatrix.times(
                     scene.camera.normalizeMatrix.times(
                       scene.camera.viewMatrix));

      // For every Model in the Scene, render the model
      // and every sub model in the Model.
      for (Model model : scene.modelList)
      {
         render_model(scene, model, w2p, fb);
      }
   }


   /**
      Recursively renderer a model without any debugging information.

      @param scene the Scene that we are rendering
      @param model the current Model object to recursively render
      @param ctm   current transformation matrix
      @param fb    FrameBuffer to hold rendered image of the model
   */
   private static void render_model(Scene scene, Model model, Matrix ctm, FrameBuffer fb)
   {
      // Get the viewport dimensions (needed for the viewport stage).
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      // Update the current transformation matrix.
      ctm = ctm.times( model.modelMatrix );

      // Render every LineSgement in the Model.
      for (LineSegment ls : model.lineSegmentList)
      {
         // Make a deep copy of the line segment.
         LineSegment ls2 = new LineSegment(ls);

         // 1. Apply the current model-to-world, world-to-view, normalize, projection transformation
         ls2.v[0].times(ctm);
         ls2.v[1].times(ctm);

         // 2. Clip the line segment to the (homogeneous) standard clip region.
         if ( Clip.clip(ls2) )
         {
            // 3. Apply the perspective division operation (clip-to-NDC coordinates).
            ls2.v[0].perspectiveDivision();
            ls2.v[1].perspectiveDivision();

            // 4. Apply the viewport transformation (NDC-to-viewport coordinates).
            ls2.v[0].x = 0.5 + w/2.001 * (ls2.v[0].x + 1.0);
            ls2.v[0].y = 0.5 + h/2.001 * (ls2.v[0].y + 1.0);
            ls2.v[1].x = 0.5 + w/2.001 * (ls2.v[1].x + 1.0);
            ls2.v[1].y = 0.5 + h/2.001 * (ls2.v[1].y + 1.0);

            // 5. Rasterize the line segment into pixels.
            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls2, fb);
            }
            else
            {
               Rasterize.rasterize(ls2, fb);
            }
         }
      }

      // Recursively render every sub model of this model.
      for (Model m : model.nestedModels)
      {
         render_model(scene, m, ctm, fb);
      }
   }



   /**
      Render a scene with quite a bit of debugging information.

      @param scene Scene object to render
      @param fb    FrameBuffer to hold rendered image of the scene
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Model in the Scene, render the model
      // and every sub model in the Model.
      for (Model model : scene.modelList)
      {
         // Create the combined world-to-view, normalize, projection transformation.
         Matrix w2p = scene.camera.projMatrix.times(
                        scene.camera.normalizeMatrix.times(
                          scene.camera.viewMatrix));

         System.out.println("=== Render Model ===================================");
         render_model_debug(scene, model, w2p, fb);
         System.out.println("=== End Model ======================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a model with quite a bit of debugging information.

      @param scene the Scene that we are rendering
      @param model the current Model object to recursively render
      @param ctm   current transformation matrix
      @param fb    FrameBuffer to hold rendered image of the model
   */
   private static void render_model_debug(Scene scene, Model model, Matrix ctm, FrameBuffer fb)
   {
      // Get the viewport dimensions (needed for the viewport stage).
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      // Update the current transformation matrix.
      ctm = ctm.times( model.modelMatrix );

      // Render every LineSgement in the Model.
      for (LineSegment ls : model.lineSegmentList)
      {
         StringBuilder result = new StringBuilder();

         // Make a deep copy of the line segment.
         LineSegment ls2 = new LineSegment(ls);

         // log interesting information to standard output
         logLineSegment("0. Model", ls2, result);

         // 1. Apply the current model-to-world, world-to-view, normalize, projection transformation
         ls2.v[0].times(ctm);
         ls2.v[1].times(ctm);

         // log interesting information to standard output
         logLineSegment("1. Clip", ls2, result);

         // 2. Clip the line segment to the (homogeneous) standard clip region.
         if ( Clip.clip(ls2) )
         {
            // log interesting information to standard output
            logLineSegment("2. Clipped", ls2, result);

            // 3. Apply the perspective division operation (clip-to-NDC coordinates).
            ls2.v[0].perspectiveDivision();
            ls2.v[1].perspectiveDivision();

            // log interesting information to standard output
            logLineSegment("3. Divided", ls2, result);

            // 4. Apply the viewport transformation (NDC-to-viewport coordinates).
            ls2.v[0].x = 0.5 + w/2.001 * (ls2.v[0].x + 1.0);
            ls2.v[0].y = 0.5 + h/2.001 * (ls2.v[0].y + 1.0);
            ls2.v[1].x = 0.5 + w/2.001 * (ls2.v[1].x + 1.0);
            ls2.v[1].y = 0.5 + h/2.001 * (ls2.v[1].y + 1.0);

            // log interesting information to standard output
            logLineSegment("4. Viewport", ls2, result);

            // 5. Rasterize the line segment into pixels.
            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls2, fb);
            }
            else
            {
               Rasterize.rasterize(ls2, fb);
            }
            if ( debug_visible ) System.out.println( result );
         }
         if ( ! debug_visible ) System.out.println();
      }

      // Recursively renderer every sub model in this model.
      for (Model m : model.nestedModels)
      {
         System.out.println("====== Render Sub Model ========================");
         render_model_debug(scene, m, ctm, fb);
         System.out.println("====== End Sub Model ===========================");
      }
   }


   /**
      This method either outputs a String representation of the given
      LineSegment or it accumulates the String representation in the
      accumulator variable (for outputting later).

      @param stage name for the pipeline stage
      @param ls LineSegment whose string representation is to be appended to the accumulator
      @param accumulator accumulates the string representations for all the line segments in this stage
   */
   private static void logLineSegment(String stage, LineSegment ls, StringBuilder accumulator)
   {
      if ( debug_visible )
         accumulator.append( stage + " " + ls.toString() );
      else
         System.out.print( stage + " " + ls.toString() );
   }
}
